import { reactive, readonly, type Ref, ref, watch } from 'vue';

export * from './2d-array';
export * from './api';

export function timeout(delay: number = 0) {
  return new Promise<void>((resolve) => setTimeout(resolve, delay));
}

export function useRefresh() {
  const key = ref(0);
  return {
    key,
    refresh() {
      return ++key.value;
    },
  };
}

type Chain<T extends Iterable<unknown>[]> = T extends never[]
  ? []
  : T extends [infer R, ...infer S extends Iterable<unknown>[]]
    ? R extends Iterable<infer U>
      ? R extends unknown[]
        ? [...R, ...Chain<S>]
        : [...U[], ...Chain<S>]
      : never
    : never;
export type Future<T = void> = PromiseWithResolvers<T>;

export function* chainIterables<const T>(iterables: Iterable<T>[]): Generator<T, void, undefined> {
  for (const iterable of iterables) {
    yield* iterable;
  }
}

export class IdPool {
  protected maxId: number;
  protected unUsedIdSet: Set<number>;

  constructor(start: number = 0) {
    this.maxId = start;
    this._usedIdSet = reactive(new Set<number>());
    this.unUsedIdSet = reactive(new Set<number>());
  }

  protected _usedIdSet: Set<number>;

  get usedIdSet() {
    return readonly(this._usedIdSet);
  }

  newId(...identifiersToRemove: number[]) {
    let id: number;
    if (this.unUsedIdSet.size) {
      [id] = this.unUsedIdSet;
      this.unUsedIdSet.delete(id);
    } else {
      id = this.maxId++;
    }
    this._usedIdSet.add(id);
    this.removeId(...identifiersToRemove);
    return id;
  }

  removeId(...identifiers: number[]) {
    for (const id of identifiers) {
      if (!this._usedIdSet.delete(id)) {
        return;
      }
      this.unUsedIdSet.add(id);
    }
  }

  reset() {
    this.maxId = 0;
    this.unUsedIdSet.clear();
    this._usedIdSet.clear();
  }
}

export class Queue<T> {
  static CancelledError = class CancelledError extends Error {};
  protected array: T[];
  protected getterFutures: Future<T>[];

  constructor(iterable: Iterable<T> = []) {
    this.array = Array.from(iterable);
    this.getterFutures = [];
  }

  get size() {
    return this.array.length;
  }

  async shift(): Promise<T> {
    if (this.array.length) {
      return this.array.shift()!;
    }
    const future: Future<T> = Promise.withResolvers();
    this.getterFutures.push(future);
    return future.promise;
  }

  push(...values: T[]): this {
    for (const value of values) {
      const future = this.getterFutures.shift();
      if (future !== undefined) {
        future.resolve(value);
      } else {
        this.array.push(value);
      }
    }
    return this;
  }

  cancelGetters() {
    for (const future of this.getterFutures) {
      future.reject(new Queue.CancelledError());
    }
  }
}

export function arrayIncludes<T, R>(array: T[], value: R): value is T & R {
  return array.includes(value as any);
}

export function waitRef<T>(ref: Ref<T>): Promise<void>;
export function waitRef<T, const R extends T>(ref: Ref<T>, ...expect: R[]): Promise<R>;
export function waitRef<T, const R extends T>(ref: Ref<T>, ...expect: R[]) {
  return new Promise<R | void>((resolve) => {
    const unWatch = watch(ref, (value) => {
      if (!expect.length) {
        unWatch();
        resolve();
      } else if (arrayIncludes(expect, value)) {
        unWatch();
        resolve(value);
      }
    });
  });
}
